@@ -3,7 +3,7 @@
 //
 // Code available from: https://verilator.org
 //
-// Copyright 2003-2024 by Wilson Snyder. This program is free software; you can
+// Copyright 2003-2025 by Wilson Snyder. This program is free software; you can
 // redistribute it and/or modify it under the terms of either the GNU
 // Lesser General Public License Version 3 or the Perl Artistic License
 // Version 2.0.
@@ -67,11 +67,21 @@
 # endif
 # define VL_LIKELY(x) __builtin_expect(!!(x), 1)  // Prefer over C++20 [[likely]]
 # define VL_UNLIKELY(x) __builtin_expect(!!(x), 0)  // Prefer over C++20 [[unlikely]]
-# define VL_UNREACHABLE __builtin_unreachable()  // C++23 std::unreachable()
 # define VL_PREFETCH_RD(p) __builtin_prefetch((p), 0)
 # define VL_PREFETCH_RW(p) __builtin_prefetch((p), 1)
 #endif
 
+#ifdef __cpp_lib_unreachable
+/// Statement that may never be reached (for coverage etc)
+# define VL_UNREACHABLE std::unreachable()  // C++23
+#elif defined(__GNUC__)
+# define VL_UNREACHABLE __builtin_unreachable()
+#elif defined(_MSC_VER)  // MSVC
+# define VL_UNREACHABLE __assume(false)
+#else
+# define VL_UNREACHABLE
+#endif
+
 // Function acquires a capability/lock (-fthread-safety)
 #define VL_ACQUIRE(...) \
         VL_CLANG_ATTR(annotate("ACQUIRE")) \
@@ -179,9 +189,6 @@
 #endif
 /// Boolean expression never hit by users (branch coverage disabled)
 # define VL_UNCOVERABLE(x) VL_UNLIKELY(x)
-#ifndef VL_UNREACHABLE
-# define VL_UNREACHABLE  ///< Statement that may never be reached (for coverage etc)
-#endif
 #ifndef VL_PREFETCH_RD
 # define VL_PREFETCH_RD(p)  ///< Prefetch pointer argument with read intent
 #endif
@@ -200,6 +207,11 @@
 
 // Comment tag that Function is pure (and thus also VL_MT_SAFE)
 #define VL_PURE VL_CLANG_ATTR(annotate("PURE"))
+// Annotated function can be called only in MT_DISABLED context, i.e. either in a code unit
+// compiled with VL_MT_DISABLED_CODE_UNIT preprocessor definition, or in the main thread.
+#define VL_MT_DISABLED \
+    VL_CLANG_ATTR(annotate("MT_DISABLED")) \
+    VL_EXCLUDES(VlOs::MtScopeMutex::s_haveThreadScope)
 // Comment tag that function is threadsafe
 #define VL_MT_SAFE VL_CLANG_ATTR(annotate("MT_SAFE"))
 // Comment tag that function is threadsafe, only if
@@ -216,7 +228,7 @@
 // protected to make sure single-caller
 #define VL_MT_UNSAFE_ONE VL_CLANG_ATTR(annotate("MT_UNSAFE_ONE"))
 // Comment tag that function is entry point of parallelization
-#define VL_MT_START VL_CLANG_ATTR(annotate("MT_START"))
+#define VL_MT_START VL_CLANG_ATTR(annotate("MT_START")) VL_REQUIRES(VlOs::MtScopeMutex::s_haveThreadScope)
 
 #ifndef VL_NO_LEGACY
 # define VL_ULL(c) (c##ULL)  // Add appropriate suffix to 64-bit constant (deprecated)
@@ -462,8 +474,7 @@ using ssize_t = uint32_t;  ///< signed size_t; returned from read()
 #ifndef VL_VALUE_STRING_MAX_WORDS
     #define VL_VALUE_STRING_MAX_WORDS 64  ///< Max size in words of String conversion operation
 #endif
-
-#define VL_VALUE_STRING_MAX_CHARS (VL_VALUE_STRING_MAX_WORDS * VL_EDATASIZE / VL_BYTESIZE)
+#define VL_VALUE_STRING_MAX_CHARS (VL_VALUE_STRING_MAX_WORDS) * 4
 
 //=========================================================================
 // Base macros
@@ -588,9 +599,9 @@ static inline double VL_ROUND(double n) {
 #endif
 
 //=========================================================================
-// Macros controlling target specific optimizations
+// Macros controlling target-specific optimizations
 
-// Define VL_PORTABLE_ONLY to disable all target specific optimizations
+// Define VL_PORTABLE_ONLY to disable all target-specific optimizations
 #ifndef VL_PORTABLE_ONLY
 # ifdef __x86_64__
 #  define VL_X86_64 1
@@ -614,9 +625,19 @@ static inline double VL_ROUND(double n) {
 //=========================================================================
 // Time and performance
 
+#include <string>
+
 namespace VlOs {
 
-extern uint64_t memUsageBytes() VL_MT_SAFE;  ///< Return memory usage in bytes, or 0 if unknown
+/// Get environment variable
+extern std::string getenvStr(const std::string& envvar,
+                             const std::string& defaultValue) VL_MT_SAFE;
+
+/// Return currently executing processor number; may do an OS call underneath so slow
+extern uint16_t getcpu() VL_MT_SAFE;
+
+/// Return memory usage in bytes, or 0 if unknown
+extern uint64_t memUsageBytes() VL_MT_SAFE;
 
 // Internal: Record CPU time, starting point on construction, and current delta from that
 class DeltaCpuTime final {
@@ -648,6 +669,14 @@ public:
         return (m_start == 0.0) ? 0.0 : gettime() - m_start;
     }
 };
+
+// Used by clang's -fthread-safety, ensures that only one instance of V3ThreadScope
+// is created at a time
+class VL_CAPABILITY("mutex") MtScopeMutex final {
+public:
+    static MtScopeMutex s_haveThreadScope;
+};
+
 }  //namespace VlOs
 
 //=========================================================================
@@ -678,7 +707,7 @@ reverse_wrapper<T> reverse_view(const T& v) {
 // Object that is returned by this function is not considered
 // as MT_SAFE and any function call on this object still
 // needs to be `VL_MT_SAFE`.
-template <class T>
+template <typename T>
 T const& as_const(T& v) VL_MT_SAFE {
     return v;
 }
